source("model_simulation.R")

library(rstan)


# model fit ---------------------------------------------------------------

fusion <- stan_model(file = "fusion.stan", model_name = "fusion", verbose = T)
fusion.vec <- stan_model(file = "fusion_vec.stan", model_name = "fusion.vec", verbose = T)
fusion.ncp <- stan_model(file = "fusion_ncp.stan", model_name = "fusion.ncp", verbose = T)

dat <- simulate_data(n.area = 20, n.sample = 50)
adapt.delta <- 0.95; iter.stan <- 2000

fit <- sampling(fusion, pars = c("beta_point","beta_area","phi","tau_sq","sigma_sq","w"), 
                data = dat$dat.fusion_old, chains = 4, cores = 4, 
                iter = iter.stan, verbose = TRUE, control = list(adapt_delta = adapt.delta)) 

fit.vec <- sampling(fusion.vec, pars = c("beta_point","beta_area","phi","tau_sq","sigma_sq","w"),
                   data = dat$dat.fusionL1, chains = 4, cores = 4, 
                   iter = iter.stan, verbose = TRUE, control = list(adapt_delta = adapt.delta)) 

fit.ncp <- sampling(fusion.ncp, pars = c("beta_point","beta_area","phi","tau_sq","sigma_sq","w","wa"),
                     data = dat$dat.fusionL1, chains = 4, cores = 4, 
                     iter = iter.stan, verbose = TRUE, control = list(adapt_delta = adapt.delta)) 

# model prediction --------------------------------------------------------

generate_y <- function(w, tau.sq, beta, X){
  # w[,i] and beta[i,] represents the ith iteration for n locations
  n <- nrow(X)
  iter <- ncol(w)
  y.pred <- matrix(0,n,iter) # n_pred by iteration matrix
  for (i in 1:iter){
    y.pred[,i] <- rnorm(n, X %*% t(beta[i,]) + w[,i], sqrt(tau.sq[i]))
  }
  return(y.pred)
}
predict_nngp_w <- function(sigma_sq, phi, pred, locs, iter, n.neighbor, w){
  
  # Helper functions
  get_index_dist <- function(pred, locs , n.neighbor, n) {
    
    i_index <- function(i, pred, locs , n.neighbor) {
      #if (is.element("fields", installed.packages()[, 1]))
      #  library(fields)
      dist <- rdist(pred[i,], locs)
      im <- sort(order(dist)[1:n.neighbor])
      return(im)
    }
    # distance matrix for location i and its neighbors 
    i_dist <- function(i, neighbor_index, pred, locs )	rdist(pred[i,],locs[neighbor_index[,i],])
    # get index of neighborhood
    neighbor_index <- sapply(1:n, i_index, pred, locs , n.neighbor)
    # get distance matrix for each i
    neighbor_dist <- sapply(1:n, i_dist, neighbor_index, pred, locs )
    return(list(i = neighbor_index, d = neighbor_dist))
  }
  get_neardistM <- function (ind, neighbor_index, locs) {
    M_i <- c(dist(locs[neighbor_index[,ind],])) # unlist the distance matrix of neighbor - neighbors
    return(M_i)
  }
  
  n <- nrow(pred)
  w.pred <- matrix(0,n,iter) # n_pred by iteration matrix
  ind_distM <- get_index_dist(pred, locs, n.neighbor, n)
  neardistM <- t(sapply(1:n, get_neardistM, ind_distM$i, locs))
  neardist <- t(ind_distM$d) 
  neardist[neardist==0] <- min(neardist[neardist!=0])/2 # set coincided location to half of minimum distance
  nearind <- t(ind_distM$i)
  Ft <- BtW <- numeric(iter)
  for (i in 1:n) { # for each predicted location
    C_nei <- diag(1.001,n.neighbor) # m by m
    h = 0;
    for (j in 1:(n.neighbor-1)){
      for (k in (j+1):n.neighbor){
        h = h + 1
        C_nei[j, k] <- exp(- neardistM[i, h])
        C_nei[k, j] <- C_nei[j, k]
      }
    }
    C_site_nei <- exp(- neardist[i, ])
    
    for (it in 1:iter){ # for each iteration
      C_site_nei_C_nei_inv <- solve(C_nei^(1/phi[it]), C_site_nei^(1/phi[it]))# m by m times m by n
      Ft[it] <- C_site_nei_C_nei_inv %*% C_site_nei^(1/phi[it])
      BtW[it] <- C_site_nei_C_nei_inv %*% w[nearind[i,],it] # 1 by m, m by 1
    }
    w.pred[i,] <- rnorm(it,BtW,sqrt(sigma_sq * (1-Ft)))
  }
  return(w.pred)
}
short_pred.fusion.w <- function(model, simulation){
  # predict with both w at observated locations and sampling locations
  fit.fusion.extract <- data.frame(rstan::extract(model))[,c("sigma_sq","tau_sq","phi")]
  fit.fusion.beta <- data.frame(rstan::extract(model,c("beta_point")))
  if (model@model_name == "fusion.ncp"){
    fit.fusion.w <- t(as.matrix(data.frame(cbind(rstan::extract(model,"w")$w, rstan::extract(model,"wa")$wa))))
  } else {
  fit.fusion.w <- t(as.matrix(data.frame(rstan::extract(model,"w")$w)))
  }
  pred.ind <- simulation$pred.ind
  iter.stan <- max(dim(fit.fusion.beta))
  pred.fusion.w <- predict_nngp_w(fit.fusion.extract$sigma_sq, fit.fusion.extract$phi, simulation$mrf[pred.ind, c("x","y")], 
                                  rbind(simulation$mrf[simulation$sample.ind, c("x","y")],simulation$dat.fusionL1$samplelocs[,c("x","y")]),
                                  iter = iter.stan, n.neighbor = simulation$dat$M, fit.fusion.w)
  return(pred.fusion.w)
}
pred_rmspe <- function(prediction){
  return(mean((dat$mrf[dat$pred.ind, "sim1"] - apply(prediction, 1, function(x) median(x, na.rm = T)))^2)^0.5)
}

# compare computational time
sum(get_elapsed_time(fit))/4; sum(get_elapsed_time(fit.vec))/4; sum(get_elapsed_time(fit.ncp))/4

# compare prediction performance
pred <- short_pred.fusion.w(fit, dat)
pred.vec <- short_pred.fusion.w(fit.vec, dat)
pred.ncp <- short_pred.fusion.w(fit.ncp, dat)
pred_rmspe(pred); pred_rmspe(pred.vec); pred_rmspe(pred.ncp)


