############################## -- 
## Efficient inference of generalizaed spatial fusion models 
## with flexible specification

## Please cite: Wang, C., and Furrer. R, (2018) Efficient inference of generalized spatial fusion models with flexible specification, Stat.

## Contains R code used to generate the simulated data

## Authors: Craig Wang, Reinhard Furrer (reinhard.furrer@math.uzh.ch)
############################## --
# n.area = number of areal regions
# dimension = dimension of the simulated region
# psill = partial sill of spatial process
# phi = effective range based on a exponential model
# nugget = 0, the nugget is incorporated into the response variance
# tau.sq = nugget
# L1 = the number of sampling points for the dataset dat.fusion
# sampling1 = sampling points strategy, default "random
# beta = p by 1 matrix of coefficents for point response
# alpha = p by 1 matrix of coefficients for areal response
# n.sample = number of locations
# n.pred = number of prediction locations
# n.neighbor = number of nearest neighbors
# f.cov = type of covariance function
# cells = number of cells used for simulating point process data (experimental)
# offset = offset values for simulating point process data (experimental)
library(gstat)
library(sp)
library(SDraw)
library(splancs)
library(broom)
library(Matrix)
library(fields)

# Main function -----------------------------------------------------------
# Simulate data for both fusion and non-fusion model according to parameters
simulate_data <- function(seed=2018, n.area=50, dimension=4000, psill=0.2, phi=300, nugget=0, tau.sq=1, L1=5, sampling1="random", beta=rbind(1,5), alpha=rbind(1,0.5), n.sample=200, n.pred=50, n.neighbor=5, f.cov="Exp", cells = 4, offset = 0.01){
  
  # Supporting functions 
  voronoipolygons = function(centroids) {
    # generate spatialpolygon
    # function adopted from http://carsonfarmer.com/2009/09/voronoi-polygons-with-r/
    z = deldir(centroids[,1], centroids[,2])
    w = tile.list(z)
    
    for (i in 1:length(w)){
      w[[i]]$x[ w[[i]]$x < 0] <- 0 
      w[[i]]$y[ w[[i]]$y < 0] <- 0 
      w[[i]]$x[ w[[i]]$x > dimension] <- dimension 
      w[[i]]$y[ w[[i]]$y > dimension] <- dimension 
    }
    
    polys = vector(mode='list', length=length(w))
    for (i in seq(along=polys)) {
      pcrds = cbind(w[[i]]$x, w[[i]]$y)
      pcrds = rbind(pcrds, pcrds[1,])
      polys[[i]] = Polygons(list(Polygon(pcrds)), ID=as.character(i))
    }
    SP = SpatialPolygons(polys)
    voronoi = SpatialPolygonsDataFrame(SP, data=data.frame(x=centroids[,1], y=centroids[,2], row.names=sapply(slot(SP, 'polygons'), function(x) slot(x, 'ID'))))
    return(voronoi)
  }
  # get the neighborhood index for each location
  # s: location matrix, each row records coordinates of one point
  # m: number of neighborhood
  i_index <- function(i, s, n.neighbor) {
    #if (is.element("fields", installed.packages()[, 1]))
    #  library(fields)
    
    if(n.neighbor >= (i - 1)) {im <- 1:(i - 1)}
    else 	{
      dist <- rdist(s[c(1,i),], s[c(1:(i-1)), ])[-1,]
      im <- sort(order(dist)[1:n.neighbor])
    }
    return(im)
  }
  # distance matrix for location i and its neighbors 
  i_dist <- function(i, neighbor_index, s)	dist(s[c(i, neighbor_index[[i - 1]]), ])
  get_index_dist <- function(s, n.neighbor) {
    n = nrow(s)
    n.neighbor = min(n.neighbor, n-1)
    # get index of neighborhood
    neighbor_index <- sapply(2:n, i_index, s, n.neighbor)
    # get distance matrix for each i
    neighbor_dist <- sapply(2:n, i_dist, neighbor_index, s)
    return(list(i = neighbor_index, d = neighbor_dist))
  }
  # get the output for c function 
  get_neardistM <- function (ind, ind_distM_d) {
    if (ind < n.neighbor ){l = ind } else {l = n.neighbor}; 
    M_i <- rep(0, n.neighbor * (n.neighbor - 1) / 2);
    if (l == 1) {}
    else{
      M_i[1: (l * (l - 1) / 2)] <- 
        c(ind_distM_d[[ind]])[(l + 1): (l * (l + 1) / 2)]
    }
    return(M_i)
  }
  get_neardist <- function (ind, ind_distM_d) {
    if (ind < n.neighbor ){l = ind } else {l = n.neighbor}; 
    D_i <- rep(0, n.neighbor);
    D_i[1:l]<-c(ind_distM_d[[ind]])[1:l]
    return( D_i)
  }
  get_nearind <- function (ind, ind_distM_i) {
    if (ind < n.neighbor ){l = ind } else {l = n.neighbor}; 
    D_i <- rep(0, n.neighbor);
    D_i[1:l]<-c(ind_distM_i[[ind]])[1:l]
    return( D_i)
  }
  
  set.seed(seed)
  
  # generate latent spatial process 
  # resolution <- dimension*2/100 # result approximate 10,000 locations
  if (f.cov=="Mat"){
    vario.model <- vgm(psill=psill,model="Mat",range=phi/sqrt(3),nugget = 0, kappa = 1.5) # see plots.R in poster
  } else {
    vario.model <- vgm(psill=psill,model="Exp",range=phi,nugget = 0)
  }
  #   xy <- data.frame(x=runif(min(n.sample*100, 100000+n.sample),0,dimension),
  #                    y=runif(min(n.sample*100, 100000+n.sample),0,dimension))
  xy.point <-  data.frame(expand.grid(seq(0, dimension, length = 200),
                                seq(0, dimension, length = 200)))
  
  e <- as(raster::extent(0, dimension, 0, dimension), "SpatialPolygons")
  grd_lrg <- as.data.frame(makegrid(e, "regular", cellsize = c(dimension/cells,dimension/cells)))
  names(grd_lrg)       <- c("x", "y")
  coordinates(grd_lrg) <- c("x", "y")
  gridded(grd_lrg) <- TRUE
  centroids_grid <- SpatialPoints(coordinates(grd_lrg))
  xy.lgcp <- centroids_grid@coords
  
  n.pts <- floor(sqrt(n.pred))
  xy.pred <- expand.grid(seq(dimension/(n.pts-1)/2, dimension-dimension/(n.pts-1)/2, length = n.pts),
                         seq(dimension/(n.pts-1)/2, dimension-dimension/(n.pts-1)/2, length = n.pts))
  names(xy.point) <- names(xy.lgcp) <- names(xy.pred) <- c("x","y")
  xy <- rbind(xy.point, xy.lgcp, xy.pred)
  
  g.dummy <- gstat(formula=z~1, locations=~x+y, dummy=T, beta=0, model=vario.model, nmax=50) # set to 50 to reduce computation time
  mrf <- predict(g.dummy, newdata=xy, nsim=1)
  
  # generate representative points for LGCP 
  
  total.area <- prod(dimension, dimension)
  c.area <- total.area/nrow(xy.lgcp)
  lgcp.lambda <- exp(mrf$sim1[(nrow(xy.point)+1):(nrow(xy.point)+nrow(xy.lgcp))])
  lgcp.coords <- spsample(e, rpois(1, total.area * max(lgcp.lambda) * offset), type = "random")
  lgcp.coords <- as.data.frame(lgcp.coords@coords[which(rbinom(length(lgcp.coords), 1, lgcp.lambda[over(lgcp.coords, grd_lrg)]/max(lgcp.lambda)) == 1), ])
  coordinates(lgcp.coords) <- ~x+y
  # how many points in each grid
  lgcp.grid <- data.frame(table(over(lgcp.coords, grd_lrg)))
  colnames(lgcp.grid) <- c("idx", "number")
  lgcp.grid <- merge(data.frame(idx = 1:length(grd_lrg)), lgcp.grid, all.x = TRUE)
  lgcp.grid$number[is.na(lgcp.grid$number)] <- 0
  
  # generate aggregated latent process -
  centroids <- data.frame(x = runif(n.area,0,dimension), y = runif(n.area,0,dimension))
  centroids$x[which.max(centroids$x)] <- centroids$y[which.max(centroids$y)] <- dimension
  centroids$x[which.min(centroids$x)] <- centroids$y[which.min(centroids$y)] <- 0
  coordinates(centroids) <- ~ x + y
  poly <- voronoi.polygons(centroids, e)
  coordinates(xy) <- ~x+y
  pips <- lapply(seq(1:n.area),function(x)  pip(xy@coords,cbind(x=poly@polygons[[x]]@Polygons[[1]]@coords[,1],y=poly@polygons[[x]]@Polygons[[1]]@coords[,2])))
  pips.index <- lapply(seq(1:n.area),function(x) inpip(xy@coords,cbind(x=poly@polygons[[x]]@Polygons[[1]]@coords[,1],y=poly@polygons[[x]]@Polygons[[1]]@coords[,2]),bound=T))
  
  aD_matrix <- matrix(0,n.area,nrow(xy@coords))
  for (i in 1:n.area){
    aD_matrix[i,pips.index[[i]]] <- 1 
  }
  aD_matrix.full <-  diag(1/rowSums(aD_matrix)) %*% aD_matrix  # aggregation matrix that computes average mrf per area 
  mean.mrf <- aD_matrix.full %*% exp(mrf$sim1) # mean(exp(w))
  
  poly.tidy <- tidy(poly)
  poly.tidy$wbar <- mean.mrf[as.numeric(poly.tidy$id)]
  
  
  #   Visual check of the simulated data
  
      # ggplot(data=poly, aes(x = long, y = lat)) +
      #   geom_tile(data=mrf[1:nrow(xy.point),],aes(x=x, y=y, fill=sim1),alpha = 1)+
      #   geom_polygon(colour = "black", fill="white",alpha=0,size = 0.3, aes(group = group)) +
      #   scale_fill_gradient2(guide = guide_colorbar(title = "value"),low = "darkblue", mid = "white", high = "darkred",midpoint =0) +
      #   geom_point(data = as.data.frame(lgcp.coords), aes(x = x, y = y)) +
      #   labs(title="Residual Risk") + theme(plot.title = element_text(hjust = 0.5))
      # ggplot(data=poly.tidy, aes(x = long, y = lat, group = group, fill = wbar)) +
      #   scale_fill_gradient2(guide = guide_colorbar(title = "Residual risk"),low = "darkblue", mid="white",high="darkred",midpoint =mean(mean.mrf))+
      #   geom_polygon(colour = "black", size = 0.3, aes(group = group))+
      #   labs(title="Aggregated residual risk") + theme(plot.title = element_text(hjust = 0.5))

  # L1 sampling locations with a centroid plus (L1-1) points within each area
  sample.locsL1 <- data.frame(do.call(rbind,sapply(1:n.area, function(i) {
      polysample <- spsample(poly@polygons[[i]],n=L1,type = sampling1)@coords
      cbind(id=i, polysample)}
      ,simplify = F)))
  colnames(sample.locsL1)[2:3] <- c("x","y")

  sample.ind <- sample(nrow(xy@coords),n.sample)
  
  # generate point response 
  Xn <- cbind(rep(1,n.sample),rnorm(n.sample,0,1))
  w <- mrf$sim1[sample.ind]
  y <- rnorm(n.sample, Xn %*% beta + w, sqrt(tau.sq))
  # generate areal response  
  Q <- numeric(n.area)
  Xa <- cbind(rep(1,n.area), exp(rnorm(n.area,-0.5,0.5)))
  
  for (i in 1:n.area){
      # mean = exp(Xm %*% alpha ) * mean(exp(w))
      Q.mean <- exp((Xa %*% alpha)[i]) * (poly.tidy$wbar[as.numeric(poly.tidy$id)==i][1]) 
      Q[i]<-rpois(1, Q.mean )
  }
  poly.tidy$Q <- Q[as.numeric(poly.tidy$id)]
  
  A0L1 <- cbind(diag(n.sample), matrix(0, n.sample, nrow(sample.locsL1)))
  A1L1.2nd <- as.matrix(bdiag(lapply(1:n.area, function(i) t(matrix(rep(1/L1, L1))))))  # the non-0 part of A1, block diag
  A1L1 <- cbind(matrix(0,n.area,n.sample),A1L1.2nd)
  
  fit <- lm(y~Xn-1)
  beta.starting <- coefficients(fit)
  alpha.starting <- coefficients(glm(Q~Xa-1,family="poisson"))
  beta.tuning <- t(chol(vcov(fit)))
  
  # NNGP 
  ind_distM <- get_index_dist(mrf[sample.ind,c("x","y")], n.neighbor)
  neardistM <- t(sapply(1: (n.sample- 1), get_neardistM, ind_distM$d))
  neardist <- t(sapply(1:(n.sample-1), get_neardist, ind_distM$d))
  nearind <- t(sapply(1: (n.sample-1), get_nearind, ind_distM$i))
  
  FCOV = ifelse(f.cov=="Exp",1,2)
  
  sC_nei_point <- array(NA, dim = c(n.sample - 1, n.neighbor, n.neighbor))
  sC_site_nei_point <- array(NA, dim = c(n.sample - 1, n.neighbor))
  
  for (i in 2:n.sample){
    dim <- if (i < (n.neighbor+1)){i-1} else {n.neighbor}
    h = 0;
    sC_nei_point[i-1,,] <- diag(0, n.neighbor) # m by m
    for (j in 1:(dim-1)){
      for (k in (j+1):dim){
        h = h + 1
        sC_nei_point[i-1, j, k] <- - neardistM[i-1, h]
        sC_nei_point[i-1, k, j] <- sC_nei_point[i-1, j, k]
      }
    }
    sC_site_nei_point[i-1,] <- - neardist[i-1, ]
  }
  
  dat <- list(n_point = n.sample, M = n.neighbor, p_point = 2, Y = y, X_point = Xn,
              sC_nei = sC_nei_point, sC_site_nei = sC_site_nei_point, nearind = nearind, FCOV = FCOV)
  
  # point and area fusion
  ind_distML1 <- get_index_dist(rbind(mrf[sample.ind,c("x","y")],sample.locsL1[,c("x","y")]), n.neighbor)
  neardistML1 <- t(sapply(1: (n.sample + nrow(sample.locsL1) - 1), get_neardistM, ind_distML1$d))
  neardistL1 <- t(sapply(1:(n.sample + nrow(sample.locsL1)-1), get_neardist, ind_distML1$d))
  nearindL1 <- t(sapply(1: (n.sample + nrow(sample.locsL1)-1), get_nearind, ind_distML1$i))
  
  # kriging based on same NNGP as fusion model
  neardistM_sample <- neardistML1[(n.sample):(n.sample+nrow(sample.locsL1)-1),]
  neardist_sample <- neardistL1[(n.sample):(n.sample+nrow(sample.locsL1)-1),]
  nearind_sample <- nearindL1[(n.sample):(n.sample+nrow(sample.locsL1)-1),]
  
  C_nei <- array(NA, dim = c(nrow(sample.locsL1), n.neighbor, n.neighbor))
  C_site_nei <- array(NA, dim = c(nrow(sample.locsL1), n.neighbor))
  
  for (i in 1:nrow(sample.locsL1)){
    h = 0;
    C_nei[i,,] <- diag(0, n.neighbor) # m by m
    for (j in 1:(n.neighbor-1)){
      for (k in (j+1):n.neighbor){
        h = h + 1
        C_nei[i, j, k] <- - neardistM_sample[i, h]
        C_nei[i, k, j] <- C_nei[i, j, k]
      }
    }
    C_site_nei[i,] <- - neardist_sample[i, ]
  }
  
  sC_nei <- array(NA, dim = c(n.sample-1, n.neighbor, n.neighbor))
  sC_site_nei <- array(NA, dim = c(n.sample-1, n.neighbor))
  
  for (i in 2:n.sample){
    dim <- if (i < (n.neighbor+1)){i-1} else {n.neighbor}
    h = 0;
    sC_nei[i-1,,] <- diag(0, n.neighbor) # m by m
    for (j in 1:(dim-1)){
      for (k in (j+1):dim){
        h = h + 1
        sC_nei[i-1, j, k] <- - neardistM[i-1, h]
        sC_nei[i-1, k, j] <- sC_nei[i-1, j, k]
      }
    }
    sC_site_nei[i-1,] <- - neardist[i-1, ]
  }
  
  aD_matrix.partial <- diag(1/rowSums(aD_matrix[,sample.ind])) %*% aD_matrix[,sample.ind] # aggregation matrix based on the observations
  aD_matrix.partial[is.nan(aD_matrix.partial)]<-0
  A1 <- A1L1[, (n.sample + 1) : (n.sample + nrow(sample.locsL1))]
  
  dat.fusionL1 <- list(n_point = n.sample, M = n.neighbor, n_area = n.area, aL = nrow(sample.locsL1), p_point = 2, p_area = 2, Y = y, X_point = Xn, X_area = Xa, C_nei = C_nei, C_site_nei = C_site_nei, sC_nei = sC_nei, sC_site_nei = sC_site_nei, nearind_sample = nearind_sample,  nearind = nearind, Q = Q, A0 = A0L1, A1 = A1, aD_matrix = aD_matrix.partial, FCOV=FCOV, samplelocs = sample.locsL1)
  
  dat.fusion_old <- list(n_point = n.sample, M = n.neighbor, n_area = n.area, aL = nrow(sample.locsL1), p_point = 2, p_area = 2, Y = y, X_point = Xn, X_area = Xa, nearind = nearindL1, neardist = neardistL1, neardistM = neardistML1, Q = Q, A0 = A0L1, A1 = A1, aD_matrix = aD_matrix.partial, FCOV=FCOV, samplelocs = sample.locsL1)
  
  # matrix for LGCP and observed points
  n_grid <- nrow(xy.lgcp)
  ind_distM_LGCP_point <- get_index_dist(rbind(mrf[sample.ind,c("x","y")], xy.lgcp[,c("x","y")]), n.neighbor)
  neardistM_LGCP_point <- t(sapply(1: (n.sample + n_grid - 1), get_neardistM, ind_distM_LGCP_point$d))
  neardist_LGCP_point <- t(sapply(1:(n.sample + n_grid - 1), get_neardist, ind_distM_LGCP_point$d))
  nearind_LGCP_point <- t(sapply(1: (n.sample + n_grid - 1), get_nearind, ind_distM_LGCP_point$i))
  
  sC_nei_LGCP_point <- array(NA, dim = c(n.sample + n_grid-1, n.neighbor, n.neighbor))
  sC_site_nei_LGCP_point <- array(NA, dim = c(n.sample + n_grid-1, n.neighbor))
  
  for (i in 2:(n.sample + n_grid)){
    dim <- if (i < (n.neighbor+1)){i-1} else {n.neighbor}
    h = 0;
    sC_nei_LGCP_point[i-1,,] <- diag(0, n.neighbor) # m by m
    for (j in 1:(dim-1)){
      for (k in (j+1):dim){
        h = h + 1
        sC_nei_LGCP_point[i-1, j, k] <- - neardistM_LGCP_point[i-1, h]
        sC_nei_LGCP_point[i-1, k, j] <- sC_nei_LGCP_point[i-1, j, k]
      }
    }
    sC_site_nei_LGCP_point[i-1,] <- - neardist_LGCP_point[i-1, ]
  }

   
   dat.fusion_LGCP_point <- list(n_point = n.sample, M = n.neighbor, n_grid = n_grid, p_point = 2, Y = y, X_point = Xn, nearind = nearind_LGCP_point, sC_nei = sC_nei_LGCP_point, sC_site_nei = sC_site_nei_LGCP_point, P = lgcp.grid$number, area = c.area)

   
# fusion all   

   ind_distM_all <- get_index_dist(rbind(mrf[sample.ind,c("x","y")], xy.lgcp[,c("x","y")], sample.locsL1[,c("x","y")]), n.neighbor)
   neardistM_all <- t(sapply(1: (n.sample + n_grid + nrow(sample.locsL1)- 1), get_neardistM, ind_distM_all$d))
   neardist_all <- t(sapply(1:(n.sample + n_grid + nrow(sample.locsL1) - 1), get_neardist, ind_distM_all$d))
   nearind_all <- t(sapply(1: (n.sample + n_grid + nrow(sample.locsL1) - 1), get_nearind, ind_distM_all$i))
   
   neardistM_sample <- neardistM_all[(n.sample+n_grid):(n.sample+n_grid+nrow(sample.locsL1)-1),]
   neardist_sample <- neardist_all[(n.sample+n_grid):(n.sample+n_grid+nrow(sample.locsL1)-1),]
   nearind_sample <- nearind_all[(n.sample+n_grid):(n.sample+n_grid+nrow(sample.locsL1)-1),]
   
   C_nei <- array(NA, dim = c(nrow(sample.locsL1), n.neighbor, n.neighbor))
   C_site_nei <- array(NA, dim = c(nrow(sample.locsL1), n.neighbor))
   
   for (i in 1:nrow(sample.locsL1)){
     h = 0;
     C_nei[i,,] <- diag(0, n.neighbor) # m by m
     for (j in 1:(n.neighbor-1)){
       for (k in (j+1):n.neighbor){
         h = h + 1
         C_nei[i, j, k] <- - neardistM_sample[i, h]
         C_nei[i, k, j] <- C_nei[i, j, k]
       }
     }
     C_site_nei[i,] <- - neardist_sample[i, ]
   }
   
   sC_nei_all <- array(NA, dim = c(n.sample + n_grid - 1, n.neighbor, n.neighbor))
   sC_site_nei_all <- array(NA, dim = c(n.sample + n_grid - 1, n.neighbor))
   
   for (i in 2:(n.sample + n_grid)){
     dim <- if (i < (n.neighbor+1)){i-1} else {n.neighbor}
     h = 0;
     sC_nei_all[i-1,,] <- diag(0, n.neighbor) # m by m
     for (j in 1:(dim-1)){
       for (k in (j+1):dim){
         h = h + 1
         sC_nei_all[i-1, j, k] <- - neardistM_all[i-1, h]
         sC_nei_all[i-1, k, j] <- sC_nei_all[i-1, j, k]
       }
     }
     sC_site_nei_all[i-1,] <- - neardist_all[i-1, ]
   }
   
   
   dat.fusion_all <- list(n_point = n.sample, M = n.neighbor, n_grid = n_grid, n_area = n.area, aL = nrow(sample.locsL1), p_point = 2, p_area = 2, 
                          Y = y, X_point = Xn, X_area = Xa, C_nei = C_nei, C_site_nei = C_site_nei, 
                          nearind = nearind_LGCP_point, nearind_sample = nearind_sample, sC_nei = sC_nei_all, sC_site_nei = sC_site_nei_all, 
                          P = lgcp.grid$number, area = c.area, lgcp.coords = lgcp.coords, 
                          A0 = A0L1, A1 = A1, aD_matrix = aD_matrix.partial, FCOV=FCOV, samplelocs = sample.locsL1, Q = Q)
   

   
    out <- list(dat=dat, dat.fusion_old = dat.fusion_old, dat.fusionL1 = dat.fusionL1, dat.fusion_LGCP_point = dat.fusion_LGCP_point, dat.fusion_all = dat.fusion_all, 
                sample.ind = sample.ind, mrf = mrf, y = y, X = Xn,
                beta.starting = beta.starting, alpha.starting = alpha.starting, beta.tuning = beta.tuning, poly = poly, mean.w = mean.mrf, 
                lgcp.grid = cbind(xy.lgcp, lgcp.grid), pred.loc = xy.pred, pred.ind = seq(nrow(xy@coords) - nrow(xy.pred) + 1,  nrow(xy@coords)))
  
  return(out)
}


# demo --------------------------------------------------------------------

## Visualize the latent process w at observed locations and aggregated latent process under log link function (e.g. Poisson areal data)
# dat <- simulate_data()
# dat$poly@data$mean.w <- c(dat$mean.w)
# dat$point <- SpatialPointsDataFrame(coords = dat$mrf[dat$sample.ind,c("x","y")], data = dat$mrf[dat$sample.ind,])
# 
# library(tmap)
# tm_shape(shp = dat$poly) +
#   tm_polygons(col = "mean.w", palette = "Blues", title = "mean(exp(w))") + 
#   tm_shape(shp = dat$point) + 
#   tm_dots(col = "sim1", size = 0.25, shape = 21, palette = "Blues", border.col = "black", title = "w") + 
#   tm_layout(frame = FALSE)



